// The sketch is auto-generated with XOD (https://xod.io).
//
// You can compile and upload it to an Arduino-compatible board with
// Arduino IDE.
//
// Rough code overview:
//
// - Configuration section
// - STL shim
// - Immutable list classes and functions
// - XOD runtime environment
// - Native node implementation
// - Program graph definition
//
// Search for comments fenced with '====' and '----' to navigate through
// the major code blocks.

#include <Arduino.h>
#include <inttypes.h>
#include <avr/pgmspace.h>
#include "XODATTiny.h"


/*=============================================================================
 *
 *
 * Configuration
 *
 *
 =============================================================================*/

#define NODE_COUNT          46
#define DEFER_NODE_COUNT    3
#define MAX_OUTPUT_COUNT    1

// Uncomment to turn on debug of the program
//#define XOD_DEBUG

// Uncomment to trace the program runtime in the Serial Monitor
//#define XOD_DEBUG_ENABLE_TRACE

/*=============================================================================
 *
 *
 * STL shim. Provides implementation for vital std::* constructs
 *
 *
 =============================================================================*/

namespace std {

template< class T > struct remove_reference      {typedef T type;};
template< class T > struct remove_reference<T&>  {typedef T type;};
template< class T > struct remove_reference<T&&> {typedef T type;};

template <class T>
typename remove_reference<T>::type&& move(T&& a) {
    return static_cast<typename remove_reference<T>::type&&>(a);
}

} // namespace std

/*=============================================================================
 *
 *
 * XOD-specific list/array implementations
 *
 *
 =============================================================================*/

#ifndef XOD_LIST_H
#define XOD_LIST_H

namespace xod {
namespace detail {

/*
 * Cursors are used internaly by iterators and list views. They are not exposed
 * directly to a list consumer.
 *
 * The base `Cursor` is an interface which provides the bare minimum of methods
 * to facilitate a single iteration pass.
 */
template<typename T> class Cursor {
  public:
    virtual ~Cursor() { }
    virtual bool isValid() const = 0;
    virtual bool value(T* out) const = 0;
    virtual void next() = 0;
};

template<typename T> class NilCursor : public Cursor<T> {
  public:
    virtual bool isValid() const { return false; }
    virtual bool value(T* out) const { return false; }
    virtual void next() { }
};

} // namespace detail

/*
 * Iterator is an object used to iterate a list once.
 *
 * Users create new iterators by calling `someList.iterate()`.
 * Iterators are created on stack and are supposed to have a
 * short live, e.g. for a duration of `for` loop or node’s
 * `evaluate` function. Iterators can’t be copied.
 *
 * Implemented as a pimpl pattern wrapper over the cursor.
 * Once created for a cursor, an iterator owns that cursor
 * and will delete the cursor object once destroyed itself.
 */
template<typename T>
class Iterator {
  public:
    static Iterator<T> nil() {
        return Iterator<T>(new detail::NilCursor<T>());
    }

    Iterator(detail::Cursor<T>* cursor)
        : _cursor(cursor)
    { }

    ~Iterator() {
        if (_cursor)
            delete _cursor;
    }

    Iterator(const Iterator& that) = delete;
    Iterator& operator=(const Iterator& that) = delete;

    Iterator(Iterator&& it)
        : _cursor(it._cursor)
    {
        it._cursor = nullptr;
    }

    Iterator& operator=(Iterator&& it) {
        auto tmp = it._cursor;
        it._cursor = _cursor;
        _cursor = tmp;
        return *this;
    }

    operator bool() const { return _cursor->isValid(); }

    bool value(T* out) const {
        return _cursor->value(out);
    }

    T operator*() const {
        T out;
        _cursor->value(&out);
        return out;
    }

    Iterator& operator++() {
        _cursor->next();
        return *this;
    }

  private:
    detail::Cursor<T>* _cursor;
};

/*
 * An interface for a list view. A particular list view provides a new
 * kind of iteration over existing data. This way we can use list slices,
 * list concatenations, list rotations, etc without introducing new data
 * buffers. We just change the way already existing data is iterated.
 *
 * ListView is not exposed to a list user directly, it is used internally
 * by the List class. However, deriving a new ListView is necessary if you
 * make a new list/string processing node.
 */
template<typename T> class ListView {
  public:
    virtual Iterator<T> iterate() const = 0;
};

/*
 * The list as it seen by data consumers. Have a single method `iterate`
 * to create a new iterator.
 *
 * Implemented as pimpl pattern wrapper over a list view. Takes pointer
 * to a list view in constructor and expects the view will be alive for
 * the whole life time of the list.
 */
template<typename T> class List {
  public:
    constexpr List()
        : _view(nullptr)
    { }

    List(const ListView<T>* view)
        : _view(view)
    { }

    Iterator<T> iterate() const {
        return _view ? _view->iterate() : Iterator<T>::nil();
    }

    // pre 0.15.0 backward compatibility
    List* operator->() __attribute__ ((deprecated)) { return this; }
    const List* operator->() const __attribute__ ((deprecated)) { return this; }

  private:
    const ListView<T>* _view;
};

/*
 * A list view over an old good plain C array.
 *
 * Expects the array will be alive for the whole life time of the
 * view.
 */
template<typename T> class PlainListView : public ListView<T> {
  public:
    class Cursor : public detail::Cursor<T> {
      public:
        Cursor(const PlainListView* owner)
            : _owner(owner)
            , _idx(0)
        { }

        bool isValid() const override {
            return _idx < _owner->_len;
        }

        bool value(T* out) const override {
            if (!isValid())
                return false;
            *out = _owner->_data[_idx];
            return true;
        }

        void next() override { ++_idx; }

      private:
        const PlainListView* _owner;
        size_t _idx;
    };

  public:
    PlainListView(const T* data, size_t len)
        : _data(data)
        , _len(len)
    { }

    virtual Iterator<T> iterate() const override {
        return Iterator<T>(new Cursor(this));
    }

  private:
    friend class Cursor;
    const T* _data;
    size_t _len;
};

/*
 * A list view over a null-terminated C-String.
 *
 * Expects the char buffer will be alive for the whole life time of the view.
 * You can use string literals as a buffer, since they are persistent for
 * the program execution time.
 */
class CStringView : public ListView<char> {
  public:
    class Cursor : public detail::Cursor<char> {
      public:
        Cursor(const char* str)
            : _ptr(str)
        { }

        bool isValid() const override {
            return (bool)*_ptr;
        }

        bool value(char* out) const override {
            *out = *_ptr;
            return (bool)*_ptr;
        }

        void next() override { ++_ptr; }

      private:
        const char* _ptr;
    };

  public:
    CStringView(const char* str = nullptr)
        : _str(str)
    { }

    CStringView& operator=(const CStringView& rhs) {
        _str = rhs._str;
        return *this;
    }

    virtual Iterator<char> iterate() const override {
        return _str ? Iterator<char>(new Cursor(_str)) : Iterator<char>::nil();
    }

  private:
    friend class Cursor;
    const char* _str;
};

/*
 * A list view over two other lists (Left and Right) which first iterates the
 * left one, and when exhausted, iterates the right one.
 *
 * Expects both Left and Right to be alive for the whole view life time.
 */
template<typename T> class ConcatListView : public ListView<T> {
  public:
    class Cursor : public detail::Cursor<T> {
      public:
        Cursor(Iterator<T>&& left, Iterator<T>&& right)
            : _left(std::move(left))
            , _right(std::move(right))
        { }

        bool isValid() const override {
            return _left || _right;
        }

        bool value(T* out) const override {
            return _left.value(out) || _right.value(out);
        }

        void next() override {
            _left ? ++_left : ++_right;
        }

      private:
        Iterator<T> _left;
        Iterator<T> _right;
    };

  public:
    ConcatListView() { }

    ConcatListView(List<T> left, List<T> right)
        : _left(left)
        , _right(right)
    { }

    ConcatListView& operator=(const ConcatListView& rhs) {
        _left = rhs._left;
        _right = rhs._right;
        return *this;
    }

    virtual Iterator<T> iterate() const override {
        return Iterator<T>(new Cursor(_left.iterate(), _right.iterate()));
    }

  private:
    friend class Cursor;
    const List<T> _left;
    const List<T> _right;
};

//----------------------------------------------------------------------------
// Text string helpers
//----------------------------------------------------------------------------

using XString = List<char>;

/*
 * List and list view in a single pack. An utility used to define constant
 * string literals in XOD.
 */
class XStringCString : public XString {
  public:
    XStringCString(const char* str)
        : XString(&_view)
        , _view(str)
    { }

  private:
    CStringView _view;
};

} // namespace xod

#endif

/*=============================================================================
 *
 *
 * Basic algorithms for XOD lists
 *
 *
 =============================================================================*/

#ifndef XOD_LIST_FUNCS_H
#define XOD_LIST_FUNCS_H



namespace xod {

/*
 * Folds a list from left. Also known as "reduce".
 */
template<typename T, typename TR>
TR foldl(List<T> xs, TR (*func)(TR, T), TR acc) {
    for (auto it = xs.iterate(); it; ++it)
        acc = func(acc, *it);
    return acc;
}

template<typename T> size_t lengthReducer(size_t len, T) {
    return len + 1;
}

/*
 * Computes length of a list.
 */
template<typename T> size_t length(List<T> xs) {
    return foldl(xs, lengthReducer<T>, (size_t)0);
}

template<typename T> T* dumpReducer(T* buff, T x) {
    *buff = x;
    return buff + 1;
}

/*
 * Copies a list content into a memory buffer.
 *
 * It is expected that `outBuff` has enough size to fit all the data.
 */
template<typename T> size_t dump(List<T> xs, T* outBuff) {
    T* buffEnd = foldl(xs, dumpReducer, outBuff);
    return buffEnd - outBuff;
}

} // namespace xod

#endif


/*=============================================================================
 *
 *
 * Runtime
 *
 *
 =============================================================================*/

//----------------------------------------------------------------------------
// Debug routines
//----------------------------------------------------------------------------
#ifndef DEBUG_SERIAL
#  define DEBUG_SERIAL Serial
#endif

#if defined(XOD_DEBUG) && defined(XOD_DEBUG_ENABLE_TRACE)
#  define XOD_TRACE(x)      { DEBUG_SERIAL.print(x); DEBUG_SERIAL.flush(); }
#  define XOD_TRACE_LN(x)   { DEBUG_SERIAL.println(x); DEBUG_SERIAL.flush(); }
#  define XOD_TRACE_F(x)    XOD_TRACE(F(x))
#  define XOD_TRACE_FLN(x)  XOD_TRACE_LN(F(x))
#else
#  define XOD_TRACE(x)
#  define XOD_TRACE_LN(x)
#  define XOD_TRACE_F(x)
#  define XOD_TRACE_FLN(x)
#endif

//----------------------------------------------------------------------------
// PGM space utilities
//----------------------------------------------------------------------------
#define pgm_read_nodeid(address) (pgm_read_word(address))

/*
 * Workaround for bugs:
 * https://github.com/arduino/ArduinoCore-sam/pull/43
 * https://github.com/arduino/ArduinoCore-samd/pull/253
 * Remove after the PRs merge
 */
#if !defined(ARDUINO_ARCH_AVR) && defined(pgm_read_ptr)
#  undef pgm_read_ptr
#  define pgm_read_ptr(addr) (*(const void **)(addr))
#endif

//----------------------------------------------------------------------------
// Compatibilities
//----------------------------------------------------------------------------

#if !defined(ARDUINO_ARCH_AVR)
/*
 * Provide dtostrf function for non-AVR platforms. Although many platforms
 * provide a stub many others do not. And the stub is based on `sprintf`
 * which doesn’t work with floating point formatters on some platforms
 * (e.g. Arduino M0).
 *
 * This is an implementation based on `fcvt` standard function. Taken here:
 * https://forum.arduino.cc/index.php?topic=368720.msg2542614#msg2542614
 */
char *dtostrf(double val, int width, unsigned int prec, char *sout) {
    int decpt, sign, reqd, pad;
    const char *s, *e;
    char *p;
    s = fcvt(val, prec, &decpt, &sign);
    if (prec == 0 && decpt == 0) {
        s = (*s < '5') ? "0" : "1";
        reqd = 1;
    } else {
        reqd = strlen(s);
        if (reqd > decpt) reqd++;
        if (decpt == 0) reqd++;
    }
    if (sign) reqd++;
    p = sout;
    e = p + reqd;
    pad = width - reqd;
    if (pad > 0) {
        e += pad;
        while (pad-- > 0) *p++ = ' ';
    }
    if (sign) *p++ = '-';
    if (decpt <= 0 && prec > 0) {
        *p++ = '0';
        *p++ = '.';
        e++;
        while ( decpt < 0 ) {
            decpt++;
            *p++ = '0';
        }
    }
    while (p < e) {
        *p++ = *s++;
        if (p == e) break;
        if (--decpt == 0) *p++ = '.';
    }
    if (width < 0) {
        pad = (reqd + width) * -1;
        while (pad-- > 0) *p++ = ' ';
    }
    *p = 0;
    return sout;
}
#endif


namespace xod {
//----------------------------------------------------------------------------
// Type definitions
//----------------------------------------------------------------------------
#define NO_NODE                 ((NodeId)-1)

typedef ATTinyNumber Number;
typedef bool Logic;

#if NODE_COUNT < 256
typedef uint8_t NodeId;
#elif NODE_COUNT < 65536
typedef uint16_t NodeId;
#else
typedef uint32_t NodeId;
#endif

/*
 * Context is a handle passed to each node `evaluate` function. Currently, it’s
 * alias for NodeId but likely will be changed in future to support list
 * lifting and other features
 */
typedef NodeId Context;

/*
 * LSB of a dirty flag shows whether a particular node is dirty or not
 * Other bits shows dirtieness of its particular outputs:
 * - 1-st bit for 0-th output
 * - 2-nd bit for 1-st output
 * - etc
 *
 * An outcome limitation is that a native node must not have more than 7 output
 * pins.
 */
typedef uint8_t DirtyFlags;

typedef unsigned long TimeMs;
typedef void (*EvalFuncPtr)(Context ctx);

/*
 * Each input stores a reference to its upstream node so that we can get values
 * on input pins. Having a direct pointer to the value is not enough because we
 * want to know dirty’ness as well. So we have to use this structure instead of
 * a pointer.
 */
struct UpstreamPinRef {
    // Upstream node ID
    NodeId nodeId;
    // Index of the upstream node’s output.
    // Use 3 bits as it just enough to store values 0..7
    uint16_t pinIndex : 3;
    // Byte offset in a storage of the upstream node where the actual pin value
    // is stored
    uint16_t storageOffset : 13;
};

/*
 * Input descriptor is a metaprogramming structure used to enforce an
 * input’s type and store its wiring data location as a zero-RAM constant.
 *
 * A specialized descriptor is required by `getValue` function. Every
 * input of every type node gets its own descriptor in generated code that
 * can be accessed as input_FOO. Where FOO is a pin identifier.
 */
template<typename ValueT_, size_t wiringOffset>
struct InputDescriptor {
    typedef ValueT_ ValueT;
    enum {
        WIRING_OFFSET = wiringOffset
    };
};

/*
 * Output descriptor serve the same purpose as InputDescriptor but for
 * outputs.
 *
 * In addition to wiring data location it keeps storage data location (where
 * actual value is stored) and own zero-based index among outputs of a particular
 * node
 */
template<typename ValueT_, size_t wiringOffset, size_t storageOffset, uint8_t index>
struct OutputDescriptor {
    typedef ValueT_ ValueT;
    enum {
        WIRING_OFFSET = wiringOffset,
        STORAGE_OFFSET = storageOffset,
        INDEX = index
    };
};

//----------------------------------------------------------------------------
// Forward declarations
//----------------------------------------------------------------------------
extern void* const g_storages[NODE_COUNT];
extern const void* const g_wiring[NODE_COUNT];
extern DirtyFlags g_dirtyFlags[NODE_COUNT];

// TODO: replace with a compact list
extern TimeMs g_schedule[NODE_COUNT];

void clearTimeout(NodeId nid);
bool isTimedOut(NodeId nid);

//----------------------------------------------------------------------------
// Engine (private API)
//----------------------------------------------------------------------------

TimeMs g_transactionTime;

void* getStoragePtr(NodeId nid, size_t offset) {
    return (uint8_t*)pgm_read_ptr(&g_storages[nid]) + offset;
}

template<typename T>
T getStorageValue(NodeId nid, size_t offset) {
    return *reinterpret_cast<T*>(getStoragePtr(nid, offset));
}

void* getWiringPgmPtr(NodeId nid, size_t offset) {
    return (uint8_t*)pgm_read_ptr(&g_wiring[nid]) + offset;
}

template<typename T>
T getWiringValue(NodeId nid, size_t offset) {
    T result;
    memcpy_P(&result, getWiringPgmPtr(nid, offset), sizeof(T));
    return result;
}

bool isOutputDirty(NodeId nid, uint8_t index) {
    return g_dirtyFlags[nid] & (1 << (index + 1));
}

bool isInputDirtyImpl(NodeId nid, size_t wiringOffset) {
    UpstreamPinRef ref = getWiringValue<UpstreamPinRef>(nid, wiringOffset);
    if (ref.nodeId == NO_NODE)
        return false;

    return isOutputDirty(ref.nodeId, ref.pinIndex);
}

template<typename InputT>
bool isInputDirty(NodeId nid) {
    return isInputDirtyImpl(nid, InputT::WIRING_OFFSET);
}

void markPinDirty(NodeId nid, uint8_t index) {
    g_dirtyFlags[nid] |= 1 << (index + 1);
}

void markNodeDirty(NodeId nid) {
    g_dirtyFlags[nid] |= 0x1;
}

bool isNodeDirty(NodeId nid) {
    return g_dirtyFlags[nid] & 0x1;
}

template<typename T>
T getOutputValueImpl(NodeId nid, size_t storageOffset) {
    return getStorageValue<T>(nid, storageOffset);
}

template<typename T>
T getInputValueImpl(NodeId nid, size_t wiringOffset) {
    UpstreamPinRef ref = getWiringValue<UpstreamPinRef>(nid, wiringOffset);
    if (ref.nodeId == NO_NODE)
        return (T)0;

    return getOutputValueImpl<T>(ref.nodeId, ref.storageOffset);
}

template<typename T>
struct always_false {
    enum { value = 0 };
};

// GetValue -- classical trick for partial function (API `xod::getValue`)
// template specialization
template<typename InputOutputT>
struct GetValue {
    static typename InputOutputT::ValueT getValue(Context ctx) {
        static_assert(
                always_false<InputOutputT>::value,
                "You should provide an input_XXX or output_YYY argument " \
                "in angle brackets of getValue"
                );

    }
};

template<typename ValueT, size_t wiringOffset>
struct GetValue<InputDescriptor<ValueT, wiringOffset>> {
    static ValueT getValue(Context ctx) {
        return getInputValueImpl<ValueT>(ctx, wiringOffset);
    }
};

template<typename ValueT, size_t wiringOffset, size_t storageOffset, uint8_t index>
struct GetValue<OutputDescriptor<ValueT, wiringOffset, storageOffset, index>> {
    static ValueT getValue(Context ctx) {
        return getOutputValueImpl<ValueT>(ctx, storageOffset);
    }
};

template<typename T>
void emitValueImpl(
        NodeId nid,
        size_t storageOffset,
        size_t wiringOffset,
        uint8_t index,
        T value) {

    // Store new value and make the node itself dirty
    T* storedValue = reinterpret_cast<T*>(getStoragePtr(nid, storageOffset));
    *storedValue = value;
    markPinDirty(nid, index);

    // Notify downstream nodes about changes
    // NB: linked nodes array is in PGM space
    const NodeId* pDownstreamNid = getWiringValue<const NodeId*>(nid, wiringOffset);
    NodeId downstreamNid = pgm_read_nodeid(pDownstreamNid);

    while (downstreamNid != NO_NODE) {
        markNodeDirty(downstreamNid);
        downstreamNid = pgm_read_nodeid(pDownstreamNid++);
    }
}

void evaluateNode(NodeId nid) {
    XOD_TRACE_F("eval #");
    XOD_TRACE_LN(nid);
    EvalFuncPtr eval = getWiringValue<EvalFuncPtr>(nid, 0);
    eval(nid);
}

void runTransaction() {
    g_transactionTime = millis();

    XOD_TRACE_F("Transaction started, t=");
    XOD_TRACE_LN(g_transactionTime);

    // defer-* nodes are always at the very bottom of the graph,
    // so no one will recieve values emitted by them.
    // We must evaluate them before everybody else
    // to give them a chance to emit values.
    for (NodeId nid = NODE_COUNT - DEFER_NODE_COUNT; nid < NODE_COUNT; ++nid) {
        if (isTimedOut(nid)) {
            evaluateNode(nid);
            // Clear node dirty flag, so it will evaluate
            // on "regular" pass only if it has a dirty input.
            // We must save dirty output flags,
            // or 'isInputDirty' will not work correctly in "downstream" nodes.
            g_dirtyFlags[nid] &= ~0x1;
            clearTimeout(nid);
        }
    }

    for (NodeId nid = 0; nid < NODE_COUNT; ++nid) {
        if (isNodeDirty(nid)) {
            evaluateNode(nid);

            // If the schedule is stale, clear timeout so that
            // the node would not be marked dirty again in idle
            if (isTimedOut(nid))
                clearTimeout(nid);
        }
    }

    // Clear dirtieness for all nodes and pins
    memset(g_dirtyFlags, 0, sizeof(g_dirtyFlags));

    XOD_TRACE_F("Transaction completed, t=");
    XOD_TRACE_LN(millis());
}

void idle() {
    // Mark timed out nodes dirty. Do not reset schedule here to give
    // a chance for a node to get a reasonable result from `isTimedOut`
    TimeMs now = millis();
    for (NodeId nid = 0; nid < NODE_COUNT; ++nid) {
        TimeMs t = g_schedule[nid];
        if (t && t < now)
            markNodeDirty(nid);
    }
}

//----------------------------------------------------------------------------
// Public API (can be used by native nodes’ `evaluate` functions)
//----------------------------------------------------------------------------

template<typename InputOutputT>
typename InputOutputT::ValueT getValue(Context ctx) {
    return GetValue<InputOutputT>::getValue(ctx);
}

template<typename OutputT>
void emitValue(NodeId nid, typename OutputT::ValueT value) {
    emitValueImpl(
            nid,
            OutputT::STORAGE_OFFSET,
            OutputT::WIRING_OFFSET,
            OutputT::INDEX,
            value);
}

TimeMs transactionTime() {
    return g_transactionTime;
}

void setTimeout(NodeId nid, TimeMs timeout) {
    g_schedule[nid] = transactionTime() + timeout;
}

void clearTimeout(NodeId nid) {
    g_schedule[nid] = 0;
}

bool isTimedOut(NodeId nid) {
    return g_schedule[nid] && g_schedule[nid] < transactionTime();
}

} // namespace xod

//----------------------------------------------------------------------------
// Entry point
//----------------------------------------------------------------------------
void setup() {
    // FIXME: looks like there is a rounding bug. Waiting for 100ms fights it
    delay(100);
#ifdef XOD_DEBUG
    DEBUG_SERIAL.begin(115200);
#endif
    XOD_TRACE_FLN("\n\nProgram started");
}

void loop() {
    xod::idle();
    xod::runTransaction();
}

/*=============================================================================
 *
 *
 * Native node implementations
 *
 *
 =============================================================================*/

namespace xod {

//-----------------------------------------------------------------------------
// //attiny_get_vcc implementation
//-----------------------------------------------------------------------------
namespace ____attiny_get_vcc {

struct State {
};

struct Storage {
    State state;
    Number output_VCC_V;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_UPD;
    const NodeId* output_VCC_V;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_UPD = InputDescriptor<Logic, offsetof(Wiring, input_UPD)>;

using output_VCC_V = OutputDescriptor<Number, offsetof(Wiring, output_VCC_V), offsetof(Storage, output_VCC_V), 0>;

//Based on code from: http://21stdigitalhome.blogspot.com/2014/10/trinket-attiny85-internal-temperature.html

int getADC() {
  ADCSRA  |=_BV(ADSC);           // Start conversion
  while((ADCSRA & _BV(ADSC)));   // Wait until conversion is finished
  auto low  = ADCL; // must read ADCL first - it then locks ADCH

  auto high = ADCH; // unlocks both
  return (high << 8) | low;
}

void setupADC()
{
  // Measure chip voltage (Vcc)
  ADCSRA &= ~(_BV(ADATE) |_BV(ADIE)); // Clear auto trigger and interrupt enable
  ADCSRA |= _BV(ADEN);            // Enable ADC
  ADMUX  = 0x0c | _BV(REFS2);         // Use Vcc as voltage reference,

                      // bandgap reference as ADC input
  delay(1);
}

void evaluate(Context ctx) {
  if (!isInputDirty<input_UPD>(ctx))
  return;

  setupADC();

  auto rawVcc = getADC(); // use next sample as initial average

  for (int i = 2; i < 100; i++) {     // calculate running average for 100 measurements
    rawVcc += (getADC() - rawVcc) / i;
  }

  ADCSRA &= ~(_BV(ADEN));        // disable ADC

  Number vcc_result = (double)(1125.0 / rawVcc); // Calculate Vcc (in Volts); 1125 = 1.1*1023

  emitValue<output_VCC_V>(ctx, vcc_result);
}

} // namespace ____attiny_get_vcc

//-----------------------------------------------------------------------------
// xod/core/gate_pulse implementation
//-----------------------------------------------------------------------------
namespace xod__core__gate_pulse {

struct State {
};

struct Storage {
    State state;
    Logic output_OUT;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_IN;
    UpstreamPinRef input_EN;
    const NodeId* output_OUT;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_IN = InputDescriptor<Logic, offsetof(Wiring, input_IN)>;
using input_EN = InputDescriptor<Logic, offsetof(Wiring, input_EN)>;

using output_OUT = OutputDescriptor<Logic, offsetof(Wiring, output_OUT), offsetof(Storage, output_OUT), 0>;

void evaluate(Context ctx) {
    if (getValue<input_EN>(ctx) && isInputDirty<input_IN>(ctx))
        emitValue<output_OUT>(ctx, true);
}

} // namespace xod__core__gate_pulse

//-----------------------------------------------------------------------------
// xod/core/subtract implementation
//-----------------------------------------------------------------------------
namespace xod__core__subtract {

struct State {
};

struct Storage {
    State state;
    Number output_DIFF;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_X;
    UpstreamPinRef input_Y;
    const NodeId* output_DIFF;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_X = InputDescriptor<Number, offsetof(Wiring, input_X)>;
using input_Y = InputDescriptor<Number, offsetof(Wiring, input_Y)>;

using output_DIFF = OutputDescriptor<Number, offsetof(Wiring, output_DIFF), offsetof(Storage, output_DIFF), 0>;

void evaluate(Context ctx) {
    auto x = getValue<input_X>(ctx);
    auto y = getValue<input_Y>(ctx);
    emitValue<output_DIFF>(ctx, x - y);
}

} // namespace xod__core__subtract

//-----------------------------------------------------------------------------
// xod/core/defer_number implementation
//-----------------------------------------------------------------------------
namespace xod__core__defer_number {

struct State {
};

struct Storage {
    State state;
    Number output_OUT;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_IN;
    const NodeId* output_OUT;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_IN = InputDescriptor<Number, offsetof(Wiring, input_IN)>;

using output_OUT = OutputDescriptor<Number, offsetof(Wiring, output_OUT), offsetof(Storage, output_OUT), 0>;

void evaluate(Context ctx) {
    if (isInputDirty<input_IN>(ctx)) { // This happens only when all nodes are evaluated
        setTimeout(ctx, 0);
        // This will not have any immediate effect, because
        // deferred nodes are at the very bottom of sorted graph.
        // We do this to just save the value for reemission
        // on deferred-only evaluation.
        emitValue<output_OUT>(ctx, getValue<input_IN>(ctx));
    } else { // deferred-only evaluation pass
        emitValue<output_OUT>(ctx, getValue<output_OUT>(ctx));
    }
}

} // namespace xod__core__defer_number

//-----------------------------------------------------------------------------
// xod/core/buffer implementation
//-----------------------------------------------------------------------------
namespace xod__core__buffer {

struct State {
};

struct Storage {
    State state;
    Number output_MEM;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_NEW;
    UpstreamPinRef input_UPD;
    const NodeId* output_MEM;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_NEW = InputDescriptor<Number, offsetof(Wiring, input_NEW)>;
using input_UPD = InputDescriptor<Logic, offsetof(Wiring, input_UPD)>;

using output_MEM = OutputDescriptor<Number, offsetof(Wiring, output_MEM), offsetof(Storage, output_MEM), 0>;

void evaluate(Context ctx) {
    if (!isInputDirty<input_UPD>(ctx))
        return;

    emitValue<output_MEM>(ctx, getValue<input_NEW>(ctx));
}

} // namespace xod__core__buffer

//-----------------------------------------------------------------------------
// xod/core/multiply implementation
//-----------------------------------------------------------------------------
namespace xod__core__multiply {

struct State {
};

struct Storage {
    State state;
    Number output_PROD;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_X;
    UpstreamPinRef input_Y;
    const NodeId* output_PROD;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_X = InputDescriptor<Number, offsetof(Wiring, input_X)>;
using input_Y = InputDescriptor<Number, offsetof(Wiring, input_Y)>;

using output_PROD = OutputDescriptor<Number, offsetof(Wiring, output_PROD), offsetof(Storage, output_PROD), 0>;

void evaluate(Context ctx) {
    auto x = getValue<input_X>(ctx);
    auto y = getValue<input_Y>(ctx);
    emitValue<output_PROD>(ctx, x * y);
}

} // namespace xod__core__multiply

//-----------------------------------------------------------------------------
// xod/core/add implementation
//-----------------------------------------------------------------------------
namespace xod__core__add {

struct State {
};

struct Storage {
    State state;
    Number output_SUM;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_X;
    UpstreamPinRef input_Y;
    const NodeId* output_SUM;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_X = InputDescriptor<Number, offsetof(Wiring, input_X)>;
using input_Y = InputDescriptor<Number, offsetof(Wiring, input_Y)>;

using output_SUM = OutputDescriptor<Number, offsetof(Wiring, output_SUM), offsetof(Storage, output_SUM), 0>;

void evaluate(Context ctx) {
    auto x = getValue<input_X>(ctx);
    auto y = getValue<input_Y>(ctx);
    emitValue<output_SUM>(ctx, x + y);
}

} // namespace xod__core__add

//-----------------------------------------------------------------------------
// xod/core/cast_boolean_to_pulse implementation
//-----------------------------------------------------------------------------
namespace xod__core__cast_boolean_to_pulse {

struct State {
  bool state = false;
};

struct Storage {
    State state;
    Logic output_OUT;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_IN;
    const NodeId* output_OUT;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_IN = InputDescriptor<Logic, offsetof(Wiring, input_IN)>;

using output_OUT = OutputDescriptor<Logic, offsetof(Wiring, output_OUT), offsetof(Storage, output_OUT), 0>;

void evaluate(Context ctx) {
    State* state = getState(ctx);
    auto newValue = getValue<input_IN>(ctx);

    if (newValue == true && state->state == false)
        emitValue<output_OUT>(ctx, 1);

    state->state = newValue;
}

} // namespace xod__core__cast_boolean_to_pulse

//-----------------------------------------------------------------------------
// xod/core/and implementation
//-----------------------------------------------------------------------------
namespace xod__core__and {

struct State {
};

struct Storage {
    State state;
    Logic output_AND;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_A;
    UpstreamPinRef input_B;
    const NodeId* output_AND;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_A = InputDescriptor<Logic, offsetof(Wiring, input_A)>;
using input_B = InputDescriptor<Logic, offsetof(Wiring, input_B)>;

using output_AND = OutputDescriptor<Logic, offsetof(Wiring, output_AND), offsetof(Storage, output_AND), 0>;

void evaluate(Context ctx) {
    auto a = getValue<input_A>(ctx);
    auto b = getValue<input_B>(ctx);
    emitValue<output_AND>(ctx, a && b);
}

} // namespace xod__core__and

//-----------------------------------------------------------------------------
// xod/core/flip_flop implementation
//-----------------------------------------------------------------------------
namespace xod__core__flip_flop {

struct State {
};

struct Storage {
    State state;
    Logic output_MEM;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_SET;
    UpstreamPinRef input_TGL;
    UpstreamPinRef input_RST;
    const NodeId* output_MEM;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_SET = InputDescriptor<Logic, offsetof(Wiring, input_SET)>;
using input_TGL = InputDescriptor<Logic, offsetof(Wiring, input_TGL)>;
using input_RST = InputDescriptor<Logic, offsetof(Wiring, input_RST)>;

using output_MEM = OutputDescriptor<Logic, offsetof(Wiring, output_MEM), offsetof(Storage, output_MEM), 0>;

void evaluate(Context ctx) {
    bool oldState = getValue<output_MEM>(ctx);
    bool newState = oldState;

    if (isInputDirty<input_TGL>(ctx)) {
        newState = !oldState;
    } else if (isInputDirty<input_SET>(ctx)) {
        newState = true;
    } else {
        newState = false;
    }

    if (newState == oldState)
        return;

    emitValue<output_MEM>(ctx, newState);
}

} // namespace xod__core__flip_flop

//-----------------------------------------------------------------------------
// xod/core/less implementation
//-----------------------------------------------------------------------------
namespace xod__core__less {

struct State {};

struct Storage {
    State state;
    Logic output_LT;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_LHS;
    UpstreamPinRef input_RHS;
    const NodeId* output_LT;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_LHS = InputDescriptor<Number, offsetof(Wiring, input_LHS)>;
using input_RHS = InputDescriptor<Number, offsetof(Wiring, input_RHS)>;

using output_LT = OutputDescriptor<Logic, offsetof(Wiring, output_LT), offsetof(Storage, output_LT), 0>;

void evaluate(Context ctx) {
    auto lhs = getValue<input_LHS>(ctx);
    auto rhs = getValue<input_RHS>(ctx);
    emitValue<output_LT>(ctx, lhs < rhs);
}

} // namespace xod__core__less

//-----------------------------------------------------------------------------
// xod/core/defer_pulse implementation
//-----------------------------------------------------------------------------
namespace xod__core__defer_pulse {

struct State {
};

struct Storage {
    State state;
    Logic output_OUT;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_IN;
    const NodeId* output_OUT;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_IN = InputDescriptor<Logic, offsetof(Wiring, input_IN)>;

using output_OUT = OutputDescriptor<Logic, offsetof(Wiring, output_OUT), offsetof(Storage, output_OUT), 0>;

void evaluate(Context ctx) {
    if (isInputDirty<input_IN>(ctx)) { // This happens only when all nodes are evaluated
        setTimeout(ctx, 0);
    } else {
        emitValue<output_OUT>(ctx, true);
    }
}

} // namespace xod__core__defer_pulse

//-----------------------------------------------------------------------------
// xod/core/greater implementation
//-----------------------------------------------------------------------------
namespace xod__core__greater {

struct State {
};

struct Storage {
    State state;
    Logic output_GT;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_LHS;
    UpstreamPinRef input_RHS;
    const NodeId* output_GT;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_LHS = InputDescriptor<Number, offsetof(Wiring, input_LHS)>;
using input_RHS = InputDescriptor<Number, offsetof(Wiring, input_RHS)>;

using output_GT = OutputDescriptor<Logic, offsetof(Wiring, output_GT), offsetof(Storage, output_GT), 0>;

void evaluate(Context ctx) {
    auto lhs = getValue<input_LHS>(ctx);
    auto rhs = getValue<input_RHS>(ctx);
    emitValue<output_GT>(ctx, lhs > rhs);
}

} // namespace xod__core__greater

//-----------------------------------------------------------------------------
// xod/core/boot implementation
//-----------------------------------------------------------------------------
namespace xod__core__boot {

struct State {
};

struct Storage {
    State state;
    Logic output_BOOT;
};

struct Wiring {
    EvalFuncPtr eval;
    const NodeId* output_BOOT;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using output_BOOT = OutputDescriptor<Logic, offsetof(Wiring, output_BOOT), offsetof(Storage, output_BOOT), 0>;

void evaluate(Context ctx) {
    emitValue<output_BOOT>(ctx, 1);
}

} // namespace xod__core__boot

//-----------------------------------------------------------------------------
// //attiny_i2c_master implementation
//-----------------------------------------------------------------------------
namespace ____attiny_i2c_master {

// --- Enter global namespace ---
}}
#include <Wire.h>
#include <FixedPoints.h>
#include <FixedPointsCommon.h>

namespace xod {
namespace ____attiny_i2c_master {
// --- Back to local namespace ---

struct State {
    bool begun = false;
};

struct Storage {
    State state;
    Logic output_DONE;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_SEND;
    UpstreamPinRef input_ADDR;
    UpstreamPinRef input_DATA_1;
    UpstreamPinRef input_DATA_2;
    UpstreamPinRef input_DATA_3;
    const NodeId* output_DONE;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_SEND = InputDescriptor<Logic, offsetof(Wiring, input_SEND)>;
using input_ADDR = InputDescriptor<Number, offsetof(Wiring, input_ADDR)>;
using input_DATA_1 = InputDescriptor<Number, offsetof(Wiring, input_DATA_1)>;
using input_DATA_2 = InputDescriptor<Number, offsetof(Wiring, input_DATA_2)>;
using input_DATA_3 = InputDescriptor<Number, offsetof(Wiring, input_DATA_3)>;

using output_DONE = OutputDescriptor<Logic, offsetof(Wiring, output_DONE), offsetof(Storage, output_DONE), 0>;

void evaluate(Context ctx) {
     if (!isInputDirty<input_SEND>(ctx))
      return;

     State* state = getState(ctx);
   if (!state->begun)
     {
        Wire.begin();
      state->begun = true;

     }

     auto addr = ::floorFixed(getValue<input_ADDR>(ctx)).getInteger();
     auto data = getValue<input_DATA_1>(ctx);

     Wire.beginTransmission(addr);

     Wire.write((uint8_t)((data.getInternal() & 0xFF00) >> 8));
     Wire.write((uint8_t)(data.getInternal() & 0x00FF));

     data = getValue<input_DATA_2>(ctx);

     Wire.write((uint8_t)((data.getInternal() & 0xFF00) >> 8));
     Wire.write((uint8_t)(data.getInternal() & 0x00FF));

     data = getValue<input_DATA_3>(ctx);

     Wire.write((uint8_t)((data.getInternal() & 0xFF00) >> 8));
     Wire.write((uint8_t)(data.getInternal() & 0x00FF));

     Wire.endTransmission();

     emitValue<output_DONE>(ctx, 1);

}

} // namespace ____attiny_i2c_master

//-----------------------------------------------------------------------------
// //attiny_analog_read implementation
//-----------------------------------------------------------------------------
namespace ____attiny_analog_read {

struct State {
};

struct Storage {
    State state;
    Number output_VAL;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_REF;
    UpstreamPinRef input_PORT;
    UpstreamPinRef input_UPD;
    const NodeId* output_VAL;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_REF = InputDescriptor<Logic, offsetof(Wiring, input_REF)>;
using input_PORT = InputDescriptor<Number, offsetof(Wiring, input_PORT)>;
using input_UPD = InputDescriptor<Logic, offsetof(Wiring, input_UPD)>;

using output_VAL = OutputDescriptor<Number, offsetof(Wiring, output_VAL), offsetof(Storage, output_VAL), 0>;

void evaluate(Context ctx) {
    //auto inValue = getValue<input_IN>(ctx);
    //emitValue<output_OUT>(ctx, inValue);
}

} // namespace ____attiny_analog_read

//-----------------------------------------------------------------------------
// //attiny_clock implementation
//-----------------------------------------------------------------------------
namespace ____attiny_clock {

// --- Enter global namespace ---
}}
#include <FixedPoints.h>
#include <FixedPointsCommon.h>

namespace xod {
namespace ____attiny_clock {
// --- Back to local namespace ---
struct State {
  TimeMs nextTrig;
};

struct Storage {
    State state;
    Logic output_TICK;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_EN;
    UpstreamPinRef input_IVAL;
    UpstreamPinRef input_RST;
    const NodeId* output_TICK;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_EN = InputDescriptor<Logic, offsetof(Wiring, input_EN)>;
using input_IVAL = InputDescriptor<Number, offsetof(Wiring, input_IVAL)>;
using input_RST = InputDescriptor<Logic, offsetof(Wiring, input_RST)>;

using output_TICK = OutputDescriptor<Logic, offsetof(Wiring, output_TICK), offsetof(Storage, output_TICK), 0>;

void evaluate(Context ctx) {
    State* state = getState(ctx);
    TimeMs tNow = transactionTime();
    TimeMs dt = (double)(getValue<input_IVAL>(ctx).getInteger()) * 1000 +

                    (getValue<input_IVAL>(ctx).getFraction() /

                    (double)(Number(0, ~0).getInternal() + 1)) * 1000;

    TimeMs tNext = tNow + dt;

    if (isInputDirty<input_RST>(ctx) || isInputDirty<input_EN>(ctx)) {
        // Handle enable/disable/reset
        if (dt <= 0 || !getValue<input_EN>(ctx)) {
            // Disable timeout loop on zero IVAL or explicit false on EN
            state->nextTrig = 0;
            clearTimeout(ctx);
        } else if (state->nextTrig < tNow || state->nextTrig > tNext) {
            // Start timeout from scratch
            state->nextTrig = tNext;
            setTimeout(ctx, dt);
        }
    }

    if (isTimedOut(ctx)) {
        emitValue<output_TICK>(ctx, 1);
        state->nextTrig = tNext;
        setTimeout(ctx, dt);
    }
}

} // namespace ____attiny_clock

//-----------------------------------------------------------------------------
// //attiny_digital_output implementation
//-----------------------------------------------------------------------------
namespace ____attiny_digital_output {

// --- Enter global namespace ---
}}
#include <FixedPoints.h>
#include <FixedPointsCommon.h>

namespace xod {
namespace ____attiny_digital_output {
// --- Back to local namespace ---
struct State {
  int configuredPort = -1;
};

struct Storage {
    State state;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_PORT;
    UpstreamPinRef input_SIG;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_PORT = InputDescriptor<Number, offsetof(Wiring, input_PORT)>;
using input_SIG = InputDescriptor<Logic, offsetof(Wiring, input_SIG)>;

void evaluate(Context ctx) {
  State* state = getState(ctx);
  const int port = ::floorFixed(getValue<input_PORT>(ctx)).getInteger();
  if (port != state->configuredPort) {
    ::pinMode(port, OUTPUT);
    // Store configured port so to avoid repeating `pinMode` call if just
    // SIG is updated
    state->configuredPort = port;
  }

  const bool val = getValue<input_SIG>(ctx);
  ::digitalWrite(port, val);
}

} // namespace ____attiny_digital_output

//-----------------------------------------------------------------------------
// //attiny_differential_adc implementation
//-----------------------------------------------------------------------------
namespace ____attiny_differential_adc {

// --- Enter global namespace ---
}}
#include <FixedPoints.h>
#include <FixedPointsCommon.h>

namespace xod {
namespace ____attiny_differential_adc {
// --- Back to local namespace ---
#if defined(REFS1)
#define ADMUX_REFS_MASK (0x03)
#else
#define ADMUX_REFS_MASK (0x01)
#endif

#if defined(MUX5)
#define ADMUX_MUX_MASK (0x3f)
#elif defined(MUX4)
#define ADMUX_MUX_MASK (0x1f)
#elif defined(MUX3)
#define ADMUX_MUX_MASK (0x0f)
#else
#define ADMUX_MUX_MASK (0x07)
#endif

struct State {
     int configuredPort = -1;
};

struct Storage {
    State state;
    Number output_VAL;
};

struct Wiring {
    EvalFuncPtr eval;
    UpstreamPinRef input_REF;
    UpstreamPinRef input_PORT;
    UpstreamPinRef input_UPD;
    const NodeId* output_VAL;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using input_REF = InputDescriptor<Logic, offsetof(Wiring, input_REF)>;
using input_PORT = InputDescriptor<Number, offsetof(Wiring, input_PORT)>;
using input_UPD = InputDescriptor<Logic, offsetof(Wiring, input_UPD)>;

using output_VAL = OutputDescriptor<Number, offsetof(Wiring, output_VAL), offsetof(Storage, output_VAL), 0>;

void evaluate(Context ctx) {
  if (!isInputDirty<input_UPD>(ctx)) return;

  State* state = getState(ctx);
  const int port = ::floorFixed(getValue<input_PORT>(ctx)).getInteger();
  if (port != state->configuredPort) {
    ::pinMode(port, INPUT);
    // Store configured port so to avoid repeating `pinMode` on
    // subsequent requests
    state->configuredPort = port;
  }

  ADCSRA &= ~(_BV(ADATE) |_BV(ADIE)); // Clear auto trigger and interrupt enable
  ADCSRA |= _BV(ADEN);                // Enable ADC
  ADMUX = ((INTERNAL & ADMUX_REFS_MASK) << REFS0) | ((port & ADMUX_MUX_MASK) << MUX0);

  ADCSRA  |=_BV(ADSC);           // Start conversion
  while((ADCSRA & _BV(ADSC)));

  auto low  = ADCL; // must read ADCL first - it then locks ADCH

  auto high = ADCH; // unlocks both

  double v_sense = (high << 8) | low;

  emitValue<output_VAL>(ctx, Number((double)( (2 * v_sense) / 1125.)));
}

} // namespace ____attiny_differential_adc

//-----------------------------------------------------------------------------
// xod/core/constant_boolean implementation
//-----------------------------------------------------------------------------
namespace xod__core__constant_boolean {

struct State {
};

struct Storage {
    State state;
    Logic output_VAL;
};

struct Wiring {
    EvalFuncPtr eval;
    const NodeId* output_VAL;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using output_VAL = OutputDescriptor<Logic, offsetof(Wiring, output_VAL), offsetof(Storage, output_VAL), 0>;

void evaluate(Context ctx) {
}

} // namespace xod__core__constant_boolean

//-----------------------------------------------------------------------------
// xod/core/constant_number implementation
//-----------------------------------------------------------------------------
namespace xod__core__constant_number {

struct State {};

struct Storage {
    State state;
    Number output_VAL;
};

struct Wiring {
    EvalFuncPtr eval;
    const NodeId* output_VAL;
};

State* getState(NodeId nid) {
    return reinterpret_cast<State*>(getStoragePtr(nid, 0));
}

using output_VAL = OutputDescriptor<Number, offsetof(Wiring, output_VAL), offsetof(Storage, output_VAL), 0>;

void evaluate(Context ctx) {
}

} // namespace xod__core__constant_number

} // namespace xod

/*=============================================================================
 *
 *
 * Program graph
 *
 *
 =============================================================================*/

namespace xod {

    //-------------------------------------------------------------------------
    // Dynamic data
    //-------------------------------------------------------------------------

    // Storage of #0 xod/core/boot
    constexpr Logic node_0_output_BOOT = false;
    xod__core__boot::Storage storage_0 = {
        { }, // state
        node_0_output_BOOT

    };

    // Storage of #1 xod/core/constant_boolean
    constexpr Logic node_1_output_VAL = true;
    xod__core__constant_boolean::Storage storage_1 = {
        { }, // state
        node_1_output_VAL

    };

    // Storage of #2 xod/core/constant_number
    constexpr Number node_2_output_VAL = 1;
    xod__core__constant_number::Storage storage_2 = {
        { }, // state
        node_2_output_VAL

    };

    // Storage of #3 xod/core/constant_number
    constexpr Number node_3_output_VAL = 0.01;
    xod__core__constant_number::Storage storage_3 = {
        { }, // state
        node_3_output_VAL

    };

    // Storage of #4 xod/core/constant_number
    constexpr Number node_4_output_VAL = 0.01;
    xod__core__constant_number::Storage storage_4 = {
        { }, // state
        node_4_output_VAL

    };

    // Storage of #5 xod/core/constant_number
    constexpr Number node_5_output_VAL = 0.5;
    xod__core__constant_number::Storage storage_5 = {
        { }, // state
        node_5_output_VAL

    };

    // Storage of #6 xod/core/constant_number
    constexpr Number node_6_output_VAL = 0.5;
    xod__core__constant_number::Storage storage_6 = {
        { }, // state
        node_6_output_VAL

    };

    // Storage of #7 xod/core/constant_number
    constexpr Number node_7_output_VAL = 12;
    xod__core__constant_number::Storage storage_7 = {
        { }, // state
        node_7_output_VAL

    };

    // Storage of #8 xod/core/constant_boolean
    constexpr Logic node_8_output_VAL = true;
    xod__core__constant_boolean::Storage storage_8 = {
        { }, // state
        node_8_output_VAL

    };

    // Storage of #9 xod/core/constant_number
    constexpr Number node_9_output_VAL = 1;
    xod__core__constant_number::Storage storage_9 = {
        { }, // state
        node_9_output_VAL

    };

    // Storage of #10 xod/core/constant_number
    constexpr Number node_10_output_VAL = 0.1;
    xod__core__constant_number::Storage storage_10 = {
        { }, // state
        node_10_output_VAL

    };

    // Storage of #11 xod/core/constant_boolean
    constexpr Logic node_11_output_VAL = true;
    xod__core__constant_boolean::Storage storage_11 = {
        { }, // state
        node_11_output_VAL

    };

    // Storage of #12 xod/core/constant_number
    constexpr Number node_12_output_VAL = 1;
    xod__core__constant_number::Storage storage_12 = {
        { }, // state
        node_12_output_VAL

    };

    // Storage of #13 xod/core/constant_number
    constexpr Number node_13_output_VAL = 0.1;
    xod__core__constant_number::Storage storage_13 = {
        { }, // state
        node_13_output_VAL

    };

    // Storage of #14 xod/core/constant_number
    constexpr Number node_14_output_VAL = 0.1;
    xod__core__constant_number::Storage storage_14 = {
        { }, // state
        node_14_output_VAL

    };

    // Storage of #15 xod/core/constant_number
    constexpr Number node_15_output_VAL = 0.01;
    xod__core__constant_number::Storage storage_15 = {
        { }, // state
        node_15_output_VAL

    };

    // Storage of #16 xod/core/constant_number
    constexpr Number node_16_output_VAL = 5;
    xod__core__constant_number::Storage storage_16 = {
        { }, // state
        node_16_output_VAL

    };

    // Storage of #17 xod/core/constant_boolean
    constexpr Logic node_17_output_VAL = true;
    xod__core__constant_boolean::Storage storage_17 = {
        { }, // state
        node_17_output_VAL

    };

    // Storage of #18 xod/core/constant_number
    constexpr Number node_18_output_VAL = 6;
    xod__core__constant_number::Storage storage_18 = {
        { }, // state
        node_18_output_VAL

    };

    // Storage of #19 xod/core/subtract
    constexpr Number node_19_output_DIFF = 0;
    xod__core__subtract::Storage storage_19 = {
        { }, // state
        node_19_output_DIFF

    };

    // Storage of #20 //attiny_clock
    constexpr Logic node_20_output_TICK = false;
    ____attiny_clock::Storage storage_20 = {
        { }, // state
        node_20_output_TICK

    };

    // Storage of #21 xod/core/subtract
    constexpr Number node_21_output_DIFF = 0;
    xod__core__subtract::Storage storage_21 = {
        { }, // state
        node_21_output_DIFF

    };

    // Storage of #22 //attiny_get_vcc
    constexpr Number node_22_output_VCC_V = 0;
    ____attiny_get_vcc::Storage storage_22 = {
        { }, // state
        node_22_output_VCC_V

    };

    // Storage of #23 xod/core/gate_pulse
    constexpr Logic node_23_output_OUT = false;
    xod__core__gate_pulse::Storage storage_23 = {
        { }, // state
        node_23_output_OUT

    };

    // Storage of #24 xod/core/buffer
    constexpr Number node_24_output_MEM = 0;
    xod__core__buffer::Storage storage_24 = {
        { }, // state
        node_24_output_MEM

    };

    // Storage of #25 xod/core/buffer
    constexpr Number node_25_output_MEM = 0;
    xod__core__buffer::Storage storage_25 = {
        { }, // state
        node_25_output_MEM

    };

    // Storage of #26 //attiny_analog_read
    constexpr Number node_26_output_VAL = 0;
    ____attiny_analog_read::Storage storage_26 = {
        { }, // state
        node_26_output_VAL

    };

    // Storage of #27 //attiny_differential_adc
    constexpr Number node_27_output_VAL = 0;
    ____attiny_differential_adc::Storage storage_27 = {
        { }, // state
        node_27_output_VAL

    };

    // Storage of #28 xod/core/multiply
    constexpr Number node_28_output_PROD = 0;
    xod__core__multiply::Storage storage_28 = {
        { }, // state
        node_28_output_PROD

    };

    // Storage of #29 xod/core/multiply
    constexpr Number node_29_output_PROD = 0;
    xod__core__multiply::Storage storage_29 = {
        { }, // state
        node_29_output_PROD

    };

    // Storage of #30 xod/core/multiply
    constexpr Number node_30_output_PROD = 0;
    xod__core__multiply::Storage storage_30 = {
        { }, // state
        node_30_output_PROD

    };

    // Storage of #31 xod/core/greater
    constexpr Logic node_31_output_GT = false;
    xod__core__greater::Storage storage_31 = {
        { }, // state
        node_31_output_GT

    };

    // Storage of #32 xod/core/add
    constexpr Number node_32_output_SUM = 0;
    xod__core__add::Storage storage_32 = {
        { }, // state
        node_32_output_SUM

    };

    // Storage of #33 //attiny_digital_output

    ____attiny_digital_output::Storage storage_33 = {
        { }, // state
    };

    // Storage of #34 xod/core/multiply
    constexpr Number node_34_output_PROD = 0;
    xod__core__multiply::Storage storage_34 = {
        { }, // state
        node_34_output_PROD

    };

    // Storage of #35 //attiny_i2c_master
    constexpr Logic node_35_output_DONE = false;
    ____attiny_i2c_master::Storage storage_35 = {
        { }, // state
        node_35_output_DONE

    };

    // Storage of #36 xod/core/add
    constexpr Number node_36_output_SUM = 0;
    xod__core__add::Storage storage_36 = {
        { }, // state
        node_36_output_SUM

    };

    // Storage of #37 xod/core/less
    constexpr Logic node_37_output_LT = false;
    xod__core__less::Storage storage_37 = {
        { }, // state
        node_37_output_LT

    };

    // Storage of #38 xod/core/greater
    constexpr Logic node_38_output_GT = false;
    xod__core__greater::Storage storage_38 = {
        { }, // state
        node_38_output_GT

    };

    // Storage of #39 xod/core/cast_boolean_to_pulse
    constexpr Logic node_39_output_OUT = false;
    xod__core__cast_boolean_to_pulse::Storage storage_39 = {
        { }, // state
        node_39_output_OUT

    };

    // Storage of #40 xod/core/flip_flop
    constexpr Logic node_40_output_MEM = false;
    xod__core__flip_flop::Storage storage_40 = {
        { }, // state
        node_40_output_MEM

    };

    // Storage of #41 xod/core/and
    constexpr Logic node_41_output_AND = false;
    xod__core__and::Storage storage_41 = {
        { }, // state
        node_41_output_AND

    };

    // Storage of #42 xod/core/cast_boolean_to_pulse
    constexpr Logic node_42_output_OUT = false;
    xod__core__cast_boolean_to_pulse::Storage storage_42 = {
        { }, // state
        node_42_output_OUT

    };

    // Storage of #43 xod/core/defer_number
    constexpr Number node_43_output_OUT = 0;
    xod__core__defer_number::Storage storage_43 = {
        { }, // state
        node_43_output_OUT

    };

    // Storage of #44 xod/core/defer_number
    constexpr Number node_44_output_OUT = 0;
    xod__core__defer_number::Storage storage_44 = {
        { }, // state
        node_44_output_OUT

    };

    // Storage of #45 xod/core/defer_pulse
    constexpr Logic node_45_output_OUT = false;
    xod__core__defer_pulse::Storage storage_45 = {
        { }, // state
        node_45_output_OUT

    };

    DirtyFlags g_dirtyFlags[NODE_COUNT] = {
        DirtyFlags(253),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(253),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(253),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(253),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(253),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(253),
        DirtyFlags(255),
        DirtyFlags(255),
        DirtyFlags(253)
    };

    TimeMs g_schedule[NODE_COUNT] = { 0 };

    //-------------------------------------------------------------------------
    // Static (immutable) data
    //-------------------------------------------------------------------------

    // Wiring of #0 xod/core/boot
    const NodeId outLinks_0_BOOT[] PROGMEM = { 40, 20, NO_NODE };
    const xod__core__boot::Wiring wiring_0 PROGMEM = {
        &xod__core__boot::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_0_BOOT // output_BOOT
    };

    // Wiring of #1 xod/core/constant_boolean
    const NodeId outLinks_1_VAL[] PROGMEM = { 23, NO_NODE };
    const xod__core__constant_boolean::Wiring wiring_1 PROGMEM = {
        &xod__core__constant_boolean::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_1_VAL // output_VAL
    };

    // Wiring of #2 xod/core/constant_number
    const NodeId outLinks_2_VAL[] PROGMEM = { 19, NO_NODE };
    const xod__core__constant_number::Wiring wiring_2 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_2_VAL // output_VAL
    };

    // Wiring of #3 xod/core/constant_number
    const NodeId outLinks_3_VAL[] PROGMEM = { 19, NO_NODE };
    const xod__core__constant_number::Wiring wiring_3 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_3_VAL // output_VAL
    };

    // Wiring of #4 xod/core/constant_number
    const NodeId outLinks_4_VAL[] PROGMEM = { 34, NO_NODE };
    const xod__core__constant_number::Wiring wiring_4 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_4_VAL // output_VAL
    };

    // Wiring of #5 xod/core/constant_number
    const NodeId outLinks_5_VAL[] PROGMEM = { 37, NO_NODE };
    const xod__core__constant_number::Wiring wiring_5 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_5_VAL // output_VAL
    };

    // Wiring of #6 xod/core/constant_number
    const NodeId outLinks_6_VAL[] PROGMEM = { 38, NO_NODE };
    const xod__core__constant_number::Wiring wiring_6 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_6_VAL // output_VAL
    };

    // Wiring of #7 xod/core/constant_number
    const NodeId outLinks_7_VAL[] PROGMEM = { 35, NO_NODE };
    const xod__core__constant_number::Wiring wiring_7 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_7_VAL // output_VAL
    };

    // Wiring of #8 xod/core/constant_boolean
    const NodeId outLinks_8_VAL[] PROGMEM = { 26, NO_NODE };
    const xod__core__constant_boolean::Wiring wiring_8 PROGMEM = {
        &xod__core__constant_boolean::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_8_VAL // output_VAL
    };

    // Wiring of #9 xod/core/constant_number
    const NodeId outLinks_9_VAL[] PROGMEM = { 26, NO_NODE };
    const xod__core__constant_number::Wiring wiring_9 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_9_VAL // output_VAL
    };

    // Wiring of #10 xod/core/constant_number
    const NodeId outLinks_10_VAL[] PROGMEM = { 20, NO_NODE };
    const xod__core__constant_number::Wiring wiring_10 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_10_VAL // output_VAL
    };

    // Wiring of #11 xod/core/constant_boolean
    const NodeId outLinks_11_VAL[] PROGMEM = { 20, NO_NODE };
    const xod__core__constant_boolean::Wiring wiring_11 PROGMEM = {
        &xod__core__constant_boolean::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_11_VAL // output_VAL
    };

    // Wiring of #12 xod/core/constant_number
    const NodeId outLinks_12_VAL[] PROGMEM = { 21, NO_NODE };
    const xod__core__constant_number::Wiring wiring_12 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_12_VAL // output_VAL
    };

    // Wiring of #13 xod/core/constant_number
    const NodeId outLinks_13_VAL[] PROGMEM = { 21, NO_NODE };
    const xod__core__constant_number::Wiring wiring_13 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_13_VAL // output_VAL
    };

    // Wiring of #14 xod/core/constant_number
    const NodeId outLinks_14_VAL[] PROGMEM = { 30, NO_NODE };
    const xod__core__constant_number::Wiring wiring_14 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_14_VAL // output_VAL
    };

    // Wiring of #15 xod/core/constant_number
    const NodeId outLinks_15_VAL[] PROGMEM = { 31, NO_NODE };
    const xod__core__constant_number::Wiring wiring_15 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_15_VAL // output_VAL
    };

    // Wiring of #16 xod/core/constant_number
    const NodeId outLinks_16_VAL[] PROGMEM = { 33, NO_NODE };
    const xod__core__constant_number::Wiring wiring_16 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_16_VAL // output_VAL
    };

    // Wiring of #17 xod/core/constant_boolean
    const NodeId outLinks_17_VAL[] PROGMEM = { 27, NO_NODE };
    const xod__core__constant_boolean::Wiring wiring_17 PROGMEM = {
        &xod__core__constant_boolean::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_17_VAL // output_VAL
    };

    // Wiring of #18 xod/core/constant_number
    const NodeId outLinks_18_VAL[] PROGMEM = { 27, NO_NODE };
    const xod__core__constant_number::Wiring wiring_18 PROGMEM = {
        &xod__core__constant_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        // outputs (NodeId list binding)
        outLinks_18_VAL // output_VAL
    };

    // Wiring of #19 xod/core/subtract
    const NodeId outLinks_19_DIFF[] PROGMEM = { 28, NO_NODE };
    const xod__core__subtract::Wiring wiring_19 PROGMEM = {
        &xod__core__subtract::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(2),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_X
        { NodeId(3),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_Y
        // outputs (NodeId list binding)
        outLinks_19_DIFF // output_DIFF
    };

    // Wiring of #20 //attiny_clock
    const NodeId outLinks_20_TICK[] PROGMEM = { 24, 25, 23, 22, NO_NODE };
    const ____attiny_clock::Wiring wiring_20 PROGMEM = {
        &____attiny_clock::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(11),
            xod__core__constant_boolean::output_VAL::INDEX,
            xod__core__constant_boolean::output_VAL::STORAGE_OFFSET }, // input_EN
        { NodeId(10),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_IVAL
        { NodeId(0),
            xod__core__boot::output_BOOT::INDEX,
            xod__core__boot::output_BOOT::STORAGE_OFFSET }, // input_RST
        // outputs (NodeId list binding)
        outLinks_20_TICK // output_TICK
    };

    // Wiring of #21 xod/core/subtract
    const NodeId outLinks_21_DIFF[] PROGMEM = { 29, NO_NODE };
    const xod__core__subtract::Wiring wiring_21 PROGMEM = {
        &xod__core__subtract::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(12),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_X
        { NodeId(13),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_Y
        // outputs (NodeId list binding)
        outLinks_21_DIFF // output_DIFF
    };

    // Wiring of #22 //attiny_get_vcc
    const NodeId outLinks_22_VCC_V[] PROGMEM = { 35, NO_NODE };
    const ____attiny_get_vcc::Wiring wiring_22 PROGMEM = {
        &____attiny_get_vcc::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(20),
            ____attiny_clock::output_TICK::INDEX,
            ____attiny_clock::output_TICK::STORAGE_OFFSET }, // input_UPD
        // outputs (NodeId list binding)
        outLinks_22_VCC_V // output_VCC_V
    };

    // Wiring of #23 xod/core/gate_pulse
    const NodeId outLinks_23_OUT[] PROGMEM = { 35, 26, 27, NO_NODE };
    const xod__core__gate_pulse::Wiring wiring_23 PROGMEM = {
        &xod__core__gate_pulse::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(20),
            ____attiny_clock::output_TICK::INDEX,
            ____attiny_clock::output_TICK::STORAGE_OFFSET }, // input_IN
        { NodeId(1),
            xod__core__constant_boolean::output_VAL::INDEX,
            xod__core__constant_boolean::output_VAL::STORAGE_OFFSET }, // input_EN
        // outputs (NodeId list binding)
        outLinks_23_OUT // output_OUT
    };

    // Wiring of #24 xod/core/buffer
    const NodeId outLinks_24_MEM[] PROGMEM = { 28, NO_NODE };
    const xod__core__buffer::Wiring wiring_24 PROGMEM = {
        &xod__core__buffer::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(44),
            xod__core__defer_number::output_OUT::INDEX,
            xod__core__defer_number::output_OUT::STORAGE_OFFSET }, // input_NEW
        { NodeId(20),
            ____attiny_clock::output_TICK::INDEX,
            ____attiny_clock::output_TICK::STORAGE_OFFSET }, // input_UPD
        // outputs (NodeId list binding)
        outLinks_24_MEM // output_MEM
    };

    // Wiring of #25 xod/core/buffer
    const NodeId outLinks_25_MEM[] PROGMEM = { 29, NO_NODE };
    const xod__core__buffer::Wiring wiring_25 PROGMEM = {
        &xod__core__buffer::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(43),
            xod__core__defer_number::output_OUT::INDEX,
            xod__core__defer_number::output_OUT::STORAGE_OFFSET }, // input_NEW
        { NodeId(20),
            ____attiny_clock::output_TICK::INDEX,
            ____attiny_clock::output_TICK::STORAGE_OFFSET }, // input_UPD
        // outputs (NodeId list binding)
        outLinks_25_MEM // output_MEM
    };

    // Wiring of #26 //attiny_analog_read
    const NodeId outLinks_26_VAL[] PROGMEM = { 30, NO_NODE };
    const ____attiny_analog_read::Wiring wiring_26 PROGMEM = {
        &____attiny_analog_read::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(8),
            xod__core__constant_boolean::output_VAL::INDEX,
            xod__core__constant_boolean::output_VAL::STORAGE_OFFSET }, // input_REF
        { NodeId(9),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_PORT
        { NodeId(23),
            xod__core__gate_pulse::output_OUT::INDEX,
            xod__core__gate_pulse::output_OUT::STORAGE_OFFSET }, // input_UPD
        // outputs (NodeId list binding)
        outLinks_26_VAL // output_VAL
    };

    // Wiring of #27 //attiny_differential_adc
    const NodeId outLinks_27_VAL[] PROGMEM = { 35, 31, NO_NODE };
    const ____attiny_differential_adc::Wiring wiring_27 PROGMEM = {
        &____attiny_differential_adc::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(17),
            xod__core__constant_boolean::output_VAL::INDEX,
            xod__core__constant_boolean::output_VAL::STORAGE_OFFSET }, // input_REF
        { NodeId(18),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_PORT
        { NodeId(23),
            xod__core__gate_pulse::output_OUT::INDEX,
            xod__core__gate_pulse::output_OUT::STORAGE_OFFSET }, // input_UPD
        // outputs (NodeId list binding)
        outLinks_27_VAL // output_VAL
    };

    // Wiring of #28 xod/core/multiply
    const NodeId outLinks_28_PROD[] PROGMEM = { 36, NO_NODE };
    const xod__core__multiply::Wiring wiring_28 PROGMEM = {
        &xod__core__multiply::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(24),
            xod__core__buffer::output_MEM::INDEX,
            xod__core__buffer::output_MEM::STORAGE_OFFSET }, // input_X
        { NodeId(19),
            xod__core__subtract::output_DIFF::INDEX,
            xod__core__subtract::output_DIFF::STORAGE_OFFSET }, // input_Y
        // outputs (NodeId list binding)
        outLinks_28_PROD // output_PROD
    };

    // Wiring of #29 xod/core/multiply
    const NodeId outLinks_29_PROD[] PROGMEM = { 32, NO_NODE };
    const xod__core__multiply::Wiring wiring_29 PROGMEM = {
        &xod__core__multiply::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(25),
            xod__core__buffer::output_MEM::INDEX,
            xod__core__buffer::output_MEM::STORAGE_OFFSET }, // input_X
        { NodeId(21),
            xod__core__subtract::output_DIFF::INDEX,
            xod__core__subtract::output_DIFF::STORAGE_OFFSET }, // input_Y
        // outputs (NodeId list binding)
        outLinks_29_PROD // output_PROD
    };

    // Wiring of #30 xod/core/multiply
    const NodeId outLinks_30_PROD[] PROGMEM = { 32, NO_NODE };
    const xod__core__multiply::Wiring wiring_30 PROGMEM = {
        &xod__core__multiply::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(26),
            ____attiny_analog_read::output_VAL::INDEX,
            ____attiny_analog_read::output_VAL::STORAGE_OFFSET }, // input_X
        { NodeId(14),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_Y
        // outputs (NodeId list binding)
        outLinks_30_PROD // output_PROD
    };

    // Wiring of #31 xod/core/greater
    const NodeId outLinks_31_GT[] PROGMEM = { 33, NO_NODE };
    const xod__core__greater::Wiring wiring_31 PROGMEM = {
        &xod__core__greater::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(27),
            ____attiny_differential_adc::output_VAL::INDEX,
            ____attiny_differential_adc::output_VAL::STORAGE_OFFSET }, // input_LHS
        { NodeId(15),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_RHS
        // outputs (NodeId list binding)
        outLinks_31_GT // output_GT
    };

    // Wiring of #32 xod/core/add
    const NodeId outLinks_32_SUM[] PROGMEM = { 34, 35, 43, NO_NODE };
    const xod__core__add::Wiring wiring_32 PROGMEM = {
        &xod__core__add::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(29),
            xod__core__multiply::output_PROD::INDEX,
            xod__core__multiply::output_PROD::STORAGE_OFFSET }, // input_X
        { NodeId(30),
            xod__core__multiply::output_PROD::INDEX,
            xod__core__multiply::output_PROD::STORAGE_OFFSET }, // input_Y
        // outputs (NodeId list binding)
        outLinks_32_SUM // output_SUM
    };

    // Wiring of #33 //attiny_digital_output
    const ____attiny_digital_output::Wiring wiring_33 PROGMEM = {
        &____attiny_digital_output::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(16),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_PORT
        { NodeId(31),
            xod__core__greater::output_GT::INDEX,
            xod__core__greater::output_GT::STORAGE_OFFSET }, // input_SIG
        // outputs (NodeId list binding)
    };

    // Wiring of #34 xod/core/multiply
    const NodeId outLinks_34_PROD[] PROGMEM = { 36, NO_NODE };
    const xod__core__multiply::Wiring wiring_34 PROGMEM = {
        &xod__core__multiply::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(32),
            xod__core__add::output_SUM::INDEX,
            xod__core__add::output_SUM::STORAGE_OFFSET }, // input_X
        { NodeId(4),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_Y
        // outputs (NodeId list binding)
        outLinks_34_PROD // output_PROD
    };

    // Wiring of #35 //attiny_i2c_master
    const NodeId outLinks_35_DONE[] PROGMEM = { NO_NODE };
    const ____attiny_i2c_master::Wiring wiring_35 PROGMEM = {
        &____attiny_i2c_master::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(23),
            xod__core__gate_pulse::output_OUT::INDEX,
            xod__core__gate_pulse::output_OUT::STORAGE_OFFSET }, // input_SEND
        { NodeId(7),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_ADDR
        { NodeId(22),
            ____attiny_get_vcc::output_VCC_V::INDEX,
            ____attiny_get_vcc::output_VCC_V::STORAGE_OFFSET }, // input_DATA_1
        { NodeId(27),
            ____attiny_differential_adc::output_VAL::INDEX,
            ____attiny_differential_adc::output_VAL::STORAGE_OFFSET }, // input_DATA_2
        { NodeId(32),
            xod__core__add::output_SUM::INDEX,
            xod__core__add::output_SUM::STORAGE_OFFSET }, // input_DATA_3
        // outputs (NodeId list binding)
        outLinks_35_DONE // output_DONE
    };

    // Wiring of #36 xod/core/add
    const NodeId outLinks_36_SUM[] PROGMEM = { 37, 38, 44, NO_NODE };
    const xod__core__add::Wiring wiring_36 PROGMEM = {
        &xod__core__add::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(28),
            xod__core__multiply::output_PROD::INDEX,
            xod__core__multiply::output_PROD::STORAGE_OFFSET }, // input_X
        { NodeId(34),
            xod__core__multiply::output_PROD::INDEX,
            xod__core__multiply::output_PROD::STORAGE_OFFSET }, // input_Y
        // outputs (NodeId list binding)
        outLinks_36_SUM // output_SUM
    };

    // Wiring of #37 xod/core/less
    const NodeId outLinks_37_LT[] PROGMEM = { 41, NO_NODE };
    const xod__core__less::Wiring wiring_37 PROGMEM = {
        &xod__core__less::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(36),
            xod__core__add::output_SUM::INDEX,
            xod__core__add::output_SUM::STORAGE_OFFSET }, // input_LHS
        { NodeId(5),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_RHS
        // outputs (NodeId list binding)
        outLinks_37_LT // output_LT
    };

    // Wiring of #38 xod/core/greater
    const NodeId outLinks_38_GT[] PROGMEM = { 39, NO_NODE };
    const xod__core__greater::Wiring wiring_38 PROGMEM = {
        &xod__core__greater::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(36),
            xod__core__add::output_SUM::INDEX,
            xod__core__add::output_SUM::STORAGE_OFFSET }, // input_LHS
        { NodeId(6),
            xod__core__constant_number::output_VAL::INDEX,
            xod__core__constant_number::output_VAL::STORAGE_OFFSET }, // input_RHS
        // outputs (NodeId list binding)
        outLinks_38_GT // output_GT
    };

    // Wiring of #39 xod/core/cast_boolean_to_pulse
    const NodeId outLinks_39_OUT[] PROGMEM = { 40, NO_NODE };
    const xod__core__cast_boolean_to_pulse::Wiring wiring_39 PROGMEM = {
        &xod__core__cast_boolean_to_pulse::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(38),
            xod__core__greater::output_GT::INDEX,
            xod__core__greater::output_GT::STORAGE_OFFSET }, // input_IN
        // outputs (NodeId list binding)
        outLinks_39_OUT // output_OUT
    };

    // Wiring of #40 xod/core/flip_flop
    const NodeId outLinks_40_MEM[] PROGMEM = { 41, NO_NODE };
    const xod__core__flip_flop::Wiring wiring_40 PROGMEM = {
        &xod__core__flip_flop::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(39),
            xod__core__cast_boolean_to_pulse::output_OUT::INDEX,
            xod__core__cast_boolean_to_pulse::output_OUT::STORAGE_OFFSET }, // input_SET
        { NodeId(45),
            xod__core__defer_pulse::output_OUT::INDEX,
            xod__core__defer_pulse::output_OUT::STORAGE_OFFSET }, // input_TGL
        { NodeId(0),
            xod__core__boot::output_BOOT::INDEX,
            xod__core__boot::output_BOOT::STORAGE_OFFSET }, // input_RST
        // outputs (NodeId list binding)
        outLinks_40_MEM // output_MEM
    };

    // Wiring of #41 xod/core/and
    const NodeId outLinks_41_AND[] PROGMEM = { 42, NO_NODE };
    const xod__core__and::Wiring wiring_41 PROGMEM = {
        &xod__core__and::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(40),
            xod__core__flip_flop::output_MEM::INDEX,
            xod__core__flip_flop::output_MEM::STORAGE_OFFSET }, // input_A
        { NodeId(37),
            xod__core__less::output_LT::INDEX,
            xod__core__less::output_LT::STORAGE_OFFSET }, // input_B
        // outputs (NodeId list binding)
        outLinks_41_AND // output_AND
    };

    // Wiring of #42 xod/core/cast_boolean_to_pulse
    const NodeId outLinks_42_OUT[] PROGMEM = { 45, NO_NODE };
    const xod__core__cast_boolean_to_pulse::Wiring wiring_42 PROGMEM = {
        &xod__core__cast_boolean_to_pulse::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(41),
            xod__core__and::output_AND::INDEX,
            xod__core__and::output_AND::STORAGE_OFFSET }, // input_IN
        // outputs (NodeId list binding)
        outLinks_42_OUT // output_OUT
    };

    // Wiring of #43 xod/core/defer_number
    const NodeId outLinks_43_OUT[] PROGMEM = { 25, NO_NODE };
    const xod__core__defer_number::Wiring wiring_43 PROGMEM = {
        &xod__core__defer_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(32),
            xod__core__add::output_SUM::INDEX,
            xod__core__add::output_SUM::STORAGE_OFFSET }, // input_IN
        // outputs (NodeId list binding)
        outLinks_43_OUT // output_OUT
    };

    // Wiring of #44 xod/core/defer_number
    const NodeId outLinks_44_OUT[] PROGMEM = { 24, NO_NODE };
    const xod__core__defer_number::Wiring wiring_44 PROGMEM = {
        &xod__core__defer_number::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(36),
            xod__core__add::output_SUM::INDEX,
            xod__core__add::output_SUM::STORAGE_OFFSET }, // input_IN
        // outputs (NodeId list binding)
        outLinks_44_OUT // output_OUT
    };

    // Wiring of #45 xod/core/defer_pulse
    const NodeId outLinks_45_OUT[] PROGMEM = { 40, NO_NODE };
    const xod__core__defer_pulse::Wiring wiring_45 PROGMEM = {
        &xod__core__defer_pulse::evaluate,
        // inputs (UpstreamPinRef’s initializers)
        { NodeId(42),
            xod__core__cast_boolean_to_pulse::output_OUT::INDEX,
            xod__core__cast_boolean_to_pulse::output_OUT::STORAGE_OFFSET }, // input_IN
        // outputs (NodeId list binding)
        outLinks_45_OUT // output_OUT
    };

    // PGM array with pointers to PGM wiring information structs
    const void* const g_wiring[NODE_COUNT] PROGMEM = {
        &wiring_0,
        &wiring_1,
        &wiring_2,
        &wiring_3,
        &wiring_4,
        &wiring_5,
        &wiring_6,
        &wiring_7,
        &wiring_8,
        &wiring_9,
        &wiring_10,
        &wiring_11,
        &wiring_12,
        &wiring_13,
        &wiring_14,
        &wiring_15,
        &wiring_16,
        &wiring_17,
        &wiring_18,
        &wiring_19,
        &wiring_20,
        &wiring_21,
        &wiring_22,
        &wiring_23,
        &wiring_24,
        &wiring_25,
        &wiring_26,
        &wiring_27,
        &wiring_28,
        &wiring_29,
        &wiring_30,
        &wiring_31,
        &wiring_32,
        &wiring_33,
        &wiring_34,
        &wiring_35,
        &wiring_36,
        &wiring_37,
        &wiring_38,
        &wiring_39,
        &wiring_40,
        &wiring_41,
        &wiring_42,
        &wiring_43,
        &wiring_44,
        &wiring_45
    };

    // PGM array with pointers to RAM-located storages
    void* const g_storages[NODE_COUNT] PROGMEM = {
        &storage_0,
        &storage_1,
        &storage_2,
        &storage_3,
        &storage_4,
        &storage_5,
        &storage_6,
        &storage_7,
        &storage_8,
        &storage_9,
        &storage_10,
        &storage_11,
        &storage_12,
        &storage_13,
        &storage_14,
        &storage_15,
        &storage_16,
        &storage_17,
        &storage_18,
        &storage_19,
        &storage_20,
        &storage_21,
        &storage_22,
        &storage_23,
        &storage_24,
        &storage_25,
        &storage_26,
        &storage_27,
        &storage_28,
        &storage_29,
        &storage_30,
        &storage_31,
        &storage_32,
        &storage_33,
        &storage_34,
        &storage_35,
        &storage_36,
        &storage_37,
        &storage_38,
        &storage_39,
        &storage_40,
        &storage_41,
        &storage_42,
        &storage_43,
        &storage_44,
        &storage_45
    };
}

